#!/bin/sh
#
#  palModuleMgr.sh
#  NI-PAL Module Manager Utility for Mac OS X
#  script used to un/install and un/load NI-PAL based kernel drivers and NI-PALSM
#  based user daemon services
#
#  Copyright 2004-2005,
#  National Instruments Corporation.
#  All rights reserved.
#
#  Originated: 05/11/2004
#


# global variables
kernelExtension=.kext
daemonExtension=.framework
userDaemonTool=nipalsm
natinstKernelDriverInstallDir=/System/Library/Extensions
systemUserDaemonBootDir=/Library/StartupItems
# using the following variables requires using quotes around them, evil spaces
natinstDir=/Library/Application\ Support/National\ Instruments
natinstNIPALDir="$natinstDir"/nipal
natinstUserDaemonInstallDir="$natinstNIPALDir"/Services
systemUserDaemonInstallDir=/Library/Frameworks
natinstNIPALKernelDriverBootDir="$natinstNIPALDir"/BootDrivers
natinstNIPALUserDaemonDemandDir="$natinstNIPALDir"/StartupItems

# set up folders that may be there if pal is installed
if [[ ! -d "$natinstUserDaemonInstallDir" ]]
then
   mkdir -p "$natinstUserDaemonInstallDir"
   chown -f root:admin "$natinstDir"
   chown -f root:admin "$natinstNIPALDir"
   chown -f root:admin "$natinstUserDaemonInstallDir"
fi

if [[ ! -d "$natinstNIPALKernelDriverBootDir" ]]
then
   mkdir "$natinstNIPALKernelDriverBootDir"
   chown -f root:admin "$natinstNIPALKernelDriverBootDir"
fi

if [[ ! -d "$natinstNIPALUserDaemonDemandDir" ]]
then
   mkdir "$natinstNIPALUserDaemonDemandDir"
   chown -f root:wheel "$natinstNIPALUserDaemonDemandDir"
   chmod 755 "$natinstNIPALUserDaemonDemandDir"
fi

if [[ ! -d $systemUserDaemonBootDir ]]
then
   mkdir $systemUserDaemonBootDir
   chown -f root:wheel $systemUserDaemonBootDir
   chmod 755 $systemUserDaemonBootDir
fi

echoUsage()
{
   echo "Usage:" > /dev/stderr
   echo "NI-PAL Module Manager Utility 1.9.0" > /dev/stderr
   echo "" > /dev/stderr
   echo "Usage: palModuleMgr [-i -tTYPE -sSTART][-dDEPENDENCIES]" > /dev/stderr
   echo "                                         [-oOTHER][-f][-c] MODULE |" > /dev/stderr
   echo "                      [-r][-oOTHERS][-c] MODULE |" > /dev/stderr
   echo "                      [-l][-oOTHERS] MODULE |" > /dev/stderr
   echo "                      [-u][-oOTHERS] MODULE |" > /dev/stderr
   echo "option:" > /dev/stderr
   echo "i install" > /dev/stderr
   echo "  t type           - can be \"kernelDriver\" or \"userDaemon\"" > /dev/stderr
   echo "  s start          - can be \"boot\" or \"demand\"" > /dev/stderr
   echo "  d dependencies   - a string of modules depended on" > /dev/stderr
   echo "                     put ',' between modules if neccessary" > /dev/stderr
   echo "  o others         - a string of miscellenous platform options" > /dev/stderr
   echo "                     <platform1>:<option1>=<value1>," > /dev/stderr
   echo "                                 <option2>=<value2>;" > /dev/stderr
   echo "                     <platform2>:<option3>=<value3>," > /dev/stderr
   echo "                                 ..." > /dev/stderr
   echo "                     ..." > /dev/stderr
   echo "  f force          - force installation" > /dev/stderr
   echo "  c copy           - copy binary from the location" > /dev/stderr
   echo "                     specified as part of MODULE to" > /dev/stderr
   echo "                     the corresponding system directory" > /dev/stderr
   echo "" > /dev/stderr
   echo "r remove" > /dev/stderr
   echo "  o others         - a string of miscellenous platform options" > /dev/stderr
   echo "                     <platform1>:<option1>=<value1>," > /dev/stderr
   echo "                                 <option2>=<value2>;" > /dev/stderr
   echo "                     <platform2>:<option3>=<value3>," > /dev/stderr
   echo "                                 ..." > /dev/stderr
   echo "                     ..." > /dev/stderr
   echo "  c copy           - delete binary from the corresponding system" > /dev/stderr
   echo "                     directory" > /dev/stderr
   echo "" > /dev/stderr
   echo "l load" > /dev/stderr
   echo "  o others         - a string of miscellenous platform options" > /dev/stderr
   echo "                     <platform1>:<option1>=<value1>," > /dev/stderr
   echo "                                 <option2>=<value2>;" > /dev/stderr
   echo "                     <platform2>:<option3>=<value3>," > /dev/stderr
   echo "                                 ..." > /dev/stderr
   echo "                     ..." > /dev/stderr
   echo "u unload" > /dev/stderr
   echo "  o others         - a string of miscellenous platform options" > /dev/stderr
   echo "                     <platform1>:<option1>=<value1>," > /dev/stderr
   echo "                                 <option2>=<value2>;" > /dev/stderr
   echo "                     <platform2>:<option3>=<value3>," > /dev/stderr
   echo "                                 ..." > /dev/stderr
   echo "                     ..." > /dev/stderr
   echo "MODULE             - the name of the module, for example:" > /dev/stderr
   echo "                     - export/targets/kernelk-unversioned.o (for install)" > /dev/stderr
   echo "                     - kernelk (for other operations)" > /dev/stderr
}

parseArguments()
{
   moduleCopy=0
   moduleForce=0

   while [[ "" != "$1" ]]
   do
      case ${1:0:2} in
         -i)moduleOperation=install;;
         -r)moduleOperation=remove;;
         -l)moduleOperation=load;;
         -u)moduleOperation=unload;;
         -c)moduleCopy=1;;
         -f)moduleForce=1;;
         -t)
            if [[ "-t" == "$1" ]]
            then
               shift
               moduleType=$1
            else
               moduleType=${1:2}
            fi
            if [[ "kernelDriver" != $moduleType ]] && [[ "userDaemon" != $moduleType ]]
            then
               echo "unknown module type: $moduleType" > /dev/stderr
               exit 1
            fi
            ;;
         -s)
            if [[ "-s" == "$1" ]]
            then
               shift
               moduleStart=$1
            else
               moduleStart=${1:2}
            fi
            if [[ "boot" != $moduleStart ]] && [[ "demand" != $moduleStart ]]
            then
               echo "unknown module start time: $moduleStart" > /dev/stderr
               exit 1
            fi
            ;;
         -d)
            if [[ "-d" == "$1" ]]
            then
               shift
               moduleDependency=$1
            else
               moduleDependency=${1:2}
            fi
            ;;
         -o)
            if [[ "-o" == "$1" ]]
            then
               shift
               moduleOption=$1
            else
               moduleOption=${1:2}
            fi
            ;;
         *)
            if [[ "-" == ${1:0:1} ]]
            then
               echo "Unrecognized Option: $1" > /dev/stderr
               echoUsage
               exit 1
            elif [[ "" != "$moduleName" ]]
            then
               echo "Received module, $moduleName, already. Cannot accept another module name: $1" > /dev/stderr
               echoUsage
               exit 1
            else
               moduleName=$1
            fi
            ;;
      esac
      shift
   done

   #process module name
   moduleName=`echo $moduleName | tr -s \/`
   moduleNameBase=${moduleName##*/}    # get rid of directory names
   moduleNameBase=${moduleNameBase%.*} # remove extension
   modulePath=${moduleName%/$moduleNameBase*}

   if [[ "$modulePath" == "$moduleNameBase" ]]
   then
      modulePath=""
   fi

   # post process modulePath to absolute path in case it is a relative path
   if [[ "$modulePath" != "" ]]
   then
      currentPwd=`pwd`
      cd $modulePath
      modulePath=`pwd`
      cd "$currentPwd"
   fi

   if [[ "" == $moduleName ]] || [[ "" == $moduleNameBase ]]
   then
      echo "module name missing!" > /dev/stderr
      echoUsage
      exit 1
   fi

   if [[ "install" == $moduleOperation ]]
   then
      if [[ "" == $moduleType ]] || [[ "" == $moduleStart ]]
      then
         echo "-i needs -tTYPE and -sSTART options" > /dev/stderr
         echoUsage
         exit 1
      fi
   else
      # determine moduleType for other options (remove, etc...)
      if [[ -d $natinstKernelDriverInstallDir/$moduleNameBase$kernelExtension ]] &&
         [[ -d "$natinstUserDaemonInstallDir"/$moduleNameBase$daemonExtension ]]
      then
         echo "$moduleOperations: ambiguous module: $moduleNameBase" > /dev/stderr
         exit 1
      fi
      if [[ ! -d $natinstKernelDriverInstallDir/$moduleNameBase$kernelExtension ]] &&
         [[ ! -d "$natinstUserDaemonInstallDir"/$moduleNameBase$daemonExtension ]]
      then
         echo "$moduleOperations: no existing module: $moduleNameBase" > /dev/stderr
         exit 1
      fi
      if [[ -d $natinstKernelDriverInstallDir/$moduleNameBase$kernelExtension ]]
      then
         moduleType="kernelDriver"
      else
         moduleType="userDaemon"
      fi
   fi

   # final global name to set up now that we know the exact service type
   if [[ "kernelDriver" == $moduleType ]]
   then
         moduleNameFinal=${moduleNameBase}$kernelExtension
   else
         moduleNameFinal=${moduleNameBase}$daemonExtension
   fi
}


### KERNEL DRIVER Functions ###

installkernelDriver()
{
   if [[ "0" == $moduleCopy ]] &&
      [[ $natinstKernelDriverInstallDir != $modulePath ]]
   then
      echo "$0 -i requires -c or should already be copied to $natinstKernelDriverInstallDir on MacOSX" > /dev/stderr
      return 1
   fi

   # if not installed to correct location already, deal with removing it first then install
   if [[ $natinstKernelDriverInstallDir != $modulePath ]]
   then
      if ! removekernelDriver
      then
         if [[ "1" == $moduleForce ]]
         then
            echo "Failed to remove kernel driver, but -f is specified, so $0 will install a new module image" > /dev/stderr
            echo "However, you need to reload the driver manually" > /dev/stderr
         else
            echo "Failed to remove kernel driver. Installation failed!" > /dev/stderr
            echo "Please specify -f to force installation" > /dev/stderr
            return 1
         fi
      fi

      # Now do the installation
      if ! cp -r $modulePath/$moduleNameFinal $natinstKernelDriverInstallDir/$moduleNameFinal
      then
         echo "cp -r $modulePath/$moduleNameFinal $natinstKernelDriverInstallDir/$moduleNameFinal failed" > /dev/stderr
         return 1
      fi
   else
      if [[ ! -d $natinstKernelDriverInstallDir/$moduleNameBase$kernelExtension ]]
      then
         echo "invalid path, $moduleName, doesn't exist" > /dev/stderr
      fi
   fi
   chown -R root:wheel $natinstKernelDriverInstallDir/$moduleNameBase$kernelExtension

   # Set up boot time loading
   if [[ "boot" == $moduleStart ]]
   then
      echo "$moduleNameBase" > "$natinstNIPALKernelDriverBootDir"/$moduleNameBase.rc
   fi

   return 0
}

testKernelDriverLoaded()
{
   kextstat | grep $1 &> /dev/null
   return $?
}

removekernelDriver()
{
   if [[ -d $natinstKernelDriverInstallDir/$moduleNameFinal ]] && [[ "" != $moduleNameBase ]]
   then
      if ! unloadkernelDriver
      then
         echo "Failed to unload kernel driver, removing the file anyway..." > /dev/stderr
         echo "Please manually unload the kernel module from memory later" > /dev/stderr
      fi

      if [[ "1" == $moduleCopy ]] && [[ -d $natinstKernelDriverInstallDir/$moduleNameFinal ]]
      then
         rm -rf $natinstKernelDriverInstallDir/$moduleNameFinal
      fi

      rm -f "$natinstNIPALKernelDriverBootDir"/$moduleNameBase.rc

      return 0
   fi

   return 0 # not installed
}

loadkernelDriver()
{
   if ! testKernelDriverLoaded $moduleNameBase
   then
      kextload $natinstKernelDriverInstallDir/$moduleNameFinal > /dev/stderr
   else
      echo "kernel driver already loaded: $moduleNameBase" > /dev/stderr
      return 1
   fi
   return $?
}

unloadkernelDriver()
{
   if testKernelDriverLoaded $moduleNameBase
   then
      kextunload $natinstKernelDriverInstallDir/$moduleNameFinal > /dev/stderr
      return $?
   fi
   return 0 # not loaded
}

### USER DAEMON FUNCTIONS ###

generateUserDaemonScript()
{
   echo "#!/bin/sh" > $1
   echo "#" >> $1
   echo "# Copyright `date "+%C%y"`" >> $1
   echo "# National Instruments Corporation." >> $1
   echo "# All rights reserved." >> $1
   echo "#" >> $1
   echo "# File: $moduleNameBase" >> $1
   echo "#" >> $1
   echo " " >> $1
   echo "if [ -f /etc/rc.common ]; then . /etc/rc.common; fi" >> $1
   echo " " >> $1
   echo "IsServiceRunning() {" >> $1
   echo "   ps -x -a -w -o \"pid command\" | grep $moduleNameBase | grep $userDaemonTool &> /dev/null" >> $1
   echo "   return \$?" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "SignalServiceStop() {" >> $1
   echo "   if ! IsServiceRunning; then return 0; fi" >> $1
   echo "   serviceProcessNumber=\`ps -x -a -w -o \"pid command\" | grep $moduleNameBase | head -n 1 | awk '{ print \$1 }'\` &> /dev/null" >> $1
   echo "   echo -n Stopping $moduleNameBase\(\$serviceProcessNumber\)" >> $1
   echo "   kill -INT \$serviceProcessNumber &> /dev/null" >> $1
   echo "   killStatus=\$?" >> $1
   echo "   if [ \$killStatus != 0 ]; then" >> $1
   echo "      echo ." >> $1
   echo "      echo Error sending INT signal to $moduleNameBase." >> $1
   echo "   else" >> $1
   echo "      stopTimeout=0" >> $1
   echo "      timeLimit=10" >> $1
   echo "      while [ \$stopTimeout -lt \$timeLimit ]; do" >> $1
   echo "         if ps axw -o pid | grep -w \$serviceProcessNumber > /dev/null; then" >> $1
   echo "            stopTimeout=\`expr \$stopTimeout + 1\`" >> $1
   echo "            echo -n ." >> $1
   echo "            sleep 1" >> $1
   echo "         else" >> $1
   echo "            echo ." >> $1
   echo "            echo $moduleNameBase stopped." >> $1
   echo "            break" >> $1
   echo "         fi" >> $1
   echo "      done" >> $1
   echo "      if [ \$stopTimeout -eq \$timeLimit ]; then" >> $1
   echo "         echo ." >> $1
   echo "         echo $moduleNameBase did not stop within \$timeLimit seconds, giving up." >> $1
   echo "      fi" >> $1
   echo "   fi" >> $1
   echo "   return \$killStatus" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "StartService() {" >> $1
   echo "   if IsServiceRunning; then return 1; fi" >> $1
   echo "   $userDaemonTool $systemUserDaemonInstallDir/$moduleNameFinal/$moduleNameBase" >> $1
   echo "   return \$?" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "StopService() {" >> $1
   echo "   SignalServiceStop" >> $1
   echo "   return \$?" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "RestartService() {" >> $1
   echo "   StopService" >> $1
   echo "   StartService" >> $1
   echo "   return \$?" >> $1
   echo "}" >> $1
   echo " " >> $1
   echo "RunService \"\$1\"" >> $1
   echo " " >> $1
}

generateUserDaemonStartupParameters()
{
   echo "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" > $1
   echo "<!DOCTYPE plist PUBLIC \"-//Apple Computer//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">" >> $1
   echo "<plist version=\"1.0\">" >> $1
   echo "<dict>" >> $1
   echo "	<key>Description</key>" >> $1
   echo "	<string>$moduleNameBase user daemon startup script</string>" >> $1
   echo "	<key>OrderPreference</key>" >> $1
   echo "	<string>Late</string>" >> $1
   echo "	<key>Provides</key>" >> $1
   echo "	<array>" >> $1
   echo "		<string>$moduleNameBase</string>" >> $1
   echo "	</array>" >> $1
   echo "	<key>Requires</key>" >> $1
   echo "	<array>" >> $1
   echo "		<string>nipal</string>" >> $1
   echo "	</array>" >> $1
   echo "</dict>" >> $1
   echo "</plist>" >> $1
}

removeuserStartupItem()
{
   if [ -L "$systemUserDaemonBootDir"/$moduleNameBase ]
   then
      rm -f "$systemUserDaemonBootDir"/$moduleNameBase
   elif [ -d "$systemUserDaemonBootDir"/$moduleNameBase ]
   then
      rm -f "$systemUserDaemonBootDir"/$moduleNameBase/$moduleNameBase
      rm -f "$systemUserDaemonBootDir"/$moduleNameBase/StartupParameters.plist
      rmdir "$systemUserDaemonBootDir"/$moduleNameBase
   fi

   if [ -L "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase ]
   then
      rm -f "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase
   elif [ -d "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase ]
   then
      rm -f "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/$moduleNameBase
      rm -f "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/StartupParameters.plist
      rmdir "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase
   fi
}

installuserDaemon()
{
   # if not installed to correct location already, deal with removing it first then install
   if [[ $systemUserDaemonInstallDir != $modulePath ]]
   then
      if ! removeuserDaemon
      then
         if [[ "1" == $moduleForce ]]
         then
            echo "Failed to remove user daemon, but -f is specified, so $0 will install a new module image" > /dev/stderr
            echo "However, you need to reload the service manually" > /dev/stderr
         else
            echo "Failed to remove user daemon. Installation failed!" > /dev/stderr
            echo "Please specify -f to force installation" > /dev/stderr
            return 1
         fi
      fi

      # Now do the installation
      if [[ "1" == $moduleCopy ]]
      then
         if ! cp -r $modulePath/$moduleNameFinal $systemUserDaemonInstallDir/$moduleNameFinal
         then
            echo "cp -r $modulePath/$moduleNameFinal $systemUserDaemonInstallDir/$moduleNameFinal failed" > /dev/stderr
            return 1
         fi
      else
         # make a soft link
         if ! ln -fs $modulePath/$moduleNameFinal $systemUserDaemonInstallDir
         then
            echo "ln -fs $modulePath/$moduleNameFinal $systemUserDaemonInstallDir failed" > /dev/stderr
            return 1
         fi
      fi
   else
      if [[ ! -d $systemUserDaemonInstallDir/$moduleNameFinal ]]
      then
         echo "invalid path, $moduleName, doesn't exist" > /dev/stderr
         return 1
      fi
   fi

   if [[ -d $systemUserDaemonInstallDir/$moduleNameFinal ]]
   then
      chown -R root:admin $systemUserDaemonInstallDir/$moduleNameFinal
   fi

   if ! ln -fs $systemUserDaemonInstallDir/$moduleNameFinal "$natinstUserDaemonInstallDir"
   then
      echo "ln -fs $systemUserDaemonInstallDir/$moduleNameFinal \"$natinstUserDaemonInstallDir\" failed" > /dev/stderr
      return 1
   fi

   # create StartupItems, if this is a on demand then we only create it inside
   # of natinstNIPALUserDaemonDemandDir, but on boot we create the actual StartupItem
   # in systemUserDaemonBootDir and link to it in natinstNIPALUserDaemonDemandDir

   # first clean up in case we are switching between boot and demand
   removeuserStartupItem

   if [[ "boot" == $moduleStart ]]
   then
      mkdir "$systemUserDaemonBootDir"/$moduleNameBase
      generateUserDaemonScript "$systemUserDaemonBootDir"/$moduleNameBase/$moduleNameBase
      chmod u+x "$systemUserDaemonBootDir"/$moduleNameBase/$moduleNameBase
      generateUserDaemonStartupParameters "$systemUserDaemonBootDir"/$moduleNameBase/StartupParameters.plist
      chown -R root:wheel "$systemUserDaemonBootDir"/$moduleNameBase
      ln -sf "$systemUserDaemonBootDir"/$moduleNameBase "$natinstNIPALUserDaemonDemandDir"
      chown root:wheel "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase
   else
      mkdir "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase
      generateUserDaemonScript "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/$moduleNameBase
      chmod u+x "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/$moduleNameBase
      generateUserDaemonStartupParameters "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/StartupParameters.plist
      chown -R root:wheel "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase
   fi

   return 0
}

removeuserDaemon()
{
   if [[ -e $systemUserDaemonInstallDir/$moduleNameFinal ]] && [[ "" != $moduleName ]]
   then
      if ! unloaduserDaemon
      then
         echo "Failed to unload user daemon, removing the file anyway..." > /dev/stderr
         echo "Please manually unload the user daemon from memory later" > /dev/stderr
      fi

      # remove the service from the system dir if a symbolic link
      if [[ -L $systemUserDaemonInstallDir/$moduleNameFinal ]]
      then
         rm -f $systemUserDaemonInstallDir/$moduleNameFinal
      else
         # remove service library if we are told to do so
         if [[ "1" == $moduleCopy ]] && [[ -d $systemUserDaemonInstallDir/$moduleNameFinal ]]
         then
            rm -rf $systemUserDaemonInstallDir/$moduleNameFinal
         fi
      fi
   fi

   rm -f "$natinstUserDaemonInstallDir"/$moduleNameFinal

   removeuserStartupItem

   return 0
}

loaduserDaemon()
{
   if [[ -f "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/$moduleNameBase ]]
   then
      "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/$moduleNameBase start
      return $?
   fi
   return 1
}

unloaduserDaemon()
{
   if [[ -f "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/$moduleNameBase ]]
   then
      "$natinstNIPALUserDaemonDemandDir"/$moduleNameBase/$moduleNameBase stop
      return $?
   fi
   return 0
}

### MAIN ###

# check root access
if [ `id -ur` != 0 ]; then
   echo "error: $0 must be run as root" > /dev/stderr
   exit 1
fi

parseArguments $*

#call operation command
$moduleOperation$moduleType
